{ Copyright 1997-2003 by Frank Heckenbach <heckenb@mi.uni-erlangen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 2.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. }

{$gnu-pascal,I+,W-,R-,W+}  { No automatic range checking, ranges are checked explicitly here. }

{$define CONCAT1(A,B) A##B}
{$define CONCAT(A,B) CONCAT1(A,B)}

{$ifdef LIE_ARITH}
{$define __(NAME) CONCAT(Lie,NAME)}
{$else}
{$define __(NAME) NAME}
{$endif}

{$ifdef USE_GMP}
{$define _G_(NAME) CONCAT(NAME,_GMP)}
{$else}
{$define _G_(NAME) NAME}
{$endif}

{$define _(NAME) _G_(__(NAME))}

{$if __GPC_RELEASE__ >= 20030303}
{$define INLINE(DECL) DECL; attribute (inline)}
{$define STATIC(NAME, TYPE) NAME: TYPE; attribute (static)}
{$else}
{$define INLINE(DECL) inline DECL}
{$define STATIC(NAME, TYPE) NAME: static TYPE}
{$endif}

unit _(Arith);

interface

uses GPC, Tools, Commentator_Pas, {$ifdef LIE_ARITH} Lie {$else} Simplicial {$endif}
     {$ifdef USE_GMP}, GMP {$endif};

type
  NormalInt = Integer attribute (Size = 32);
  LargeInt = Integer attribute (Size = 64);

const
  { Make sure that a LargeInt can hold a product of two NormalInt's.
    If not, compiling will stop here with a "division by zero" error. }
  Assert1 = 1 div Ord (BitSizeOf (LargeInt) >= 2 * BitSizeOf (NormalInt));

{$ifdef USE_GMP}

type
  TVV_GMP = record
  case Boolean of
    False: ({ Must "match" mpz_t }
            Alloc: Integer;
            Val: NormalInt);
    True:  (GMPVal: mpz_t)
  end;

const
  MaxVal = 1 shl (BitSizeOf (NormalInt) - 2) - 1;

{$else}

type
  TVV = Integer;

const
  { Maxium value of vector elements }
  MaxVV = 1 shl (BitSizeOf (TVV) - 1) - 1;

  { Square root of MaxVV, rounded down }
  {$if False}
  SqRtMaxVV = Trunc (SqRt (MaxVV));
  {$else}
  DummyRoot = CompilerAssert (not Odd (BitSizeOf (TVV) - 2));
  SqRtMaxVV = (1 shl (BitSizeOf (TVV) div 2 - 1)) * 362 div 256 - 1;
  {$endif}

{$endif}

type
  { Type for the values of vector elements }
  PVectorValue = ^TVectorValue;
  TVectorValue = _G_(TVV);

  { Type for an entry in a sparse vector }
  PComponent = ^TComponent;
  TComponent = record
    Index: TCount;
    Val: TVectorValue
  end;

  { Type for the elements of a sparse vector }
  PComponents = ^TComponents;
  TComponents = array [0 .. MaxVarSize div SizeOf (TComponent) - 1] of TComponent;

  { Type for a sparse vector, consisting of the elements, their current
    number and the number of entries for which memory was reserved. }
  PSparseVector = ^TSparseVector;
  TSparseVector = record
    Components: PComponents;
    MaxEl: TVectorValue;
    Count, Size: TCount
  end;
  _G_(TSV) = TSparseVector;

  { Type for a sparse matrix, stored as a list of sparse vektors.
    There is also a field to store the pivot column of each row
    during Gaussian elimination. }
  PPSparseMatrix = ^PSparseMatrix;
  PSparseMatrix = ^TSparseMatrix;
  TSparseMatrix = record
    Next: PSparseMatrix;
    Row: TSparseVector;
    PivotColumn: TCount
  end;

  { Type for the value of torsions and their multiplicity }
  PPTorsionList = ^PTorsionList;
  PTorsionList = ^TTorsionList;
  TTorsionList = record
    Next: PTorsionList;
    TVal: TVectorValue;
    Count: TCount
  end;

  { Type for a table of homology groups }
  _G_(THomTable) = array[Dim0] of record
    Betti: TCount;
    Torsion: PTorsionList
  end;

function _(GetNextRow) (const BigFaces, SmallFaces: TFaces; var NextRow: TSparseVector; MinSize: TCount; var RowCount, RowDoneCount: TCount; procedure StatusProc): Boolean;
procedure _(SmithNormalForm) (const BigFaces, SmallFaces: TFaces; MaxRank, Dim: TCount;
                              Mode: Integer; const Coeff: TVectorValue;
                              var PivotCount: TCount; var TorsionList: PTorsionList);
procedure _(WriteStrTorsion) (var s: String; TorsionList: PTorsionList; Flag: Boolean);
procedure _(WriteHomology) (var OutFile: Text; const Homology: _G_(THomTable); MinDim, MaxDim: TCount);
function _(HomologyCompute) (MinDim, MaxDim: TCount; const Facets: TFacets; HomReduced: Integer) = Res: TString;

implementation

{ Initial size for the current vector (increased when necessary) }
const
  Size0 = $200;

{ Assign maximum elements, not necessary in GMP version }
INLINE (procedure SetMax (var MaxEl1: TVectorValue; const MaxEl2: TVectorValue));
begin
  {$ifdef USE_GMP}
  var Dummy: Integer;
  Dummy := MaxEl1.Alloc;
  Dummy := MaxEl2.Alloc;
  {$else}
  MaxEl1 := MaxEl2
  {$endif}
end;

{$ifdef USE_GMP}

{$define GetMax(SparseVector)}
{$define CheckMax(MaxEl, Vector, OtherMax)}

{$else}

{ Overflow error }
procedure CoeffError;
begin
  Abort ('Coefficient overflow! Please use the GMP algorithm.')
end;

{ Check for addition overflow }
INLINE (function CheckAdd (const s1, s2: TVectorValue): TVectorValue);
begin
  if Abs (s1) > MaxVV - Abs (s2) then CoeffError;
  CheckAdd := s1 + s2
end;

{ Check for multiplication overflow }
INLINE (function CheckMult (const f1, f2: TVectorValue): TVectorValue);
begin
  if ((Abs (f1) > SqRtMaxVV) or (Abs (f2) > SqRtMaxVV)) and_then (Abs (f1) > MaxVV div Abs (f2)) then CoeffError;
  CheckMult := f1 * f2
end;

{ Find maximum absolute value in a sparse vector }
INLINE (procedure GetMax (var SparseVector: TSparseVector));
var
  i: TCount;
  m: TVectorValue;
begin
  m := 0;
  for i := 0 to SparseVector.Count - 1 do
    m := Max (m, Abs (SparseVector.Components^[i].Val));
  SparseVector.MaxEl := m
end;

{ Check for possible vector addition overflow }
INLINE (procedure CheckMax (var MaxEl: TVectorValue; Vector: PSparseVector; const OtherMax: TVectorValue));
begin
  if MaxEl > MaxVV - OtherMax then
    begin
      if Vector <> nil then
        begin
          GetMax (Vector^);
          MaxEl := Vector^.MaxEl
        end;
      if MaxEl > MaxVV - OtherMax then CoeffError
    end;
  Inc (MaxEl, OtherMax)
end;

{$endif}

{ Basic integer operations, implemented differently for machine integers and GMP types }

{$ifdef USE_GMP}
INLINE (procedure V_Init (var Dest: TVectorValue));
begin
  Dest.Alloc := -1
end;

INLINE (procedure V_Done (var Dest: TVectorValue));
begin
  if Dest.Alloc >= 0 then mpz_clear (Dest.GMPVal)
end;
{$else}
{$define V_Init(Dest)}
{$define V_Done(Dest)}
{$endif}

INLINE (procedure Set_Int (var Dest: TVectorValue; Src: NormalInt));
begin
  {$ifdef USE_GMP}
  V_Done (Dest);
  Dest.Alloc := -1;
  Dest.Val := Src
  {$else}
  Dest := Src
  {$endif}
end;

{$ifdef USE_GMP}

INLINE (procedure G_Init (var Dest: TVectorValue));
begin
  if Dest.Alloc < 0 then mpz_init (Dest.GMPVal)
end;

INLINE (procedure G_CheckInt (var Dest: TVectorValue));
begin
  if (Dest.Alloc >= 0) and (mpz_cmp_si (Dest.GMPVal, MaxVal) <= 0)
                       and (mpz_cmp_si (Dest.GMPVal, -MaxVal) >= 0) then
    Set_Int (Dest, mpz_get_si (Dest.GMPVal))
end;

{$endif}

{ The following routines must, depending on `$ifdef USE_GMP' either
  work either machine integers or with a type which at run time is
  either a machine integer or a GMP type. Important: All routines
  must also work if several parameters are identical (typically Dest
  and a Src parameter), i.e. they may not modify Dest as long as
  they need the values of parameters. }

INLINE (procedure Init_V (var Dest: TVectorValue; const Src: TVectorValue));
begin
  {$ifdef USE_GMP}
  if Src.Alloc < 0 then
    begin
      Dest.Alloc := -1;
      Dest.Val := Src.Val
    end
  else
    mpz_init_set (Dest.GMPVal, Src.GMPVal)
  {$else}
  Dest := Src
  {$endif}
end;

INLINE (procedure Set_V (var Dest: TVectorValue; const Src: TVectorValue));
begin
  {$ifdef USE_GMP}
  if Src.Alloc < 0 then
    Set_Int (Dest, Src.Val)
  else if Dest.Alloc < 0 then
    mpz_init_set (Dest.GMPVal, Src.GMPVal)
  else
    mpz_set (Dest.GMPVal, Src.GMPVal)
  {$else}
  Dest := Src
  {$endif}
end;

INLINE (procedure Abs_V (var Dest: TVectorValue; const Src: TVectorValue));
begin
  {$ifdef USE_GMP}
  if Src.Alloc < 0 then
    Set_Int (Dest, Abs (Src.Val))
  else
    begin
      G_Init (Dest);
      mpz_abs (Dest.GMPVal, Src.GMPVal)
    end
  {$else}
  Dest := Abs (Src)
  {$endif}
end;

INLINE (function eq_VI (const Val1: TVectorValue; Val2: NormalInt): Boolean);
begin
  {$ifdef USE_GMP}
  if Val1.Alloc < 0 then
    eq_VI := Val1.Val = Val2
  else
    eq_VI := mpz_cmp_si (Val1.GMPVal, Val2) = 0
  {$else}
  eq_VI := Val1 = Val2
  {$endif}
end;

INLINE (function eq_VV (const Val1, Val2: TVectorValue): Boolean);
begin
  {$ifdef USE_GMP}
  if Val1.Alloc < 0 then
    if Val2.Alloc < 0 then
      eq_VV := Val1.Val = Val2.Val
    else
      eq_VV := mpz_cmp_si (Val2.GMPVal, Val1.Val) = 0
  else
    if Val2.Alloc < 0 then
      eq_VV := mpz_cmp_si (Val1.GMPVal, Val2.Val) = 0
    else
      eq_VV := mpz_cmp (Val1.GMPVal, Val2.GMPVal) = 0
  {$else}
  eq_VV := Val1 = Val2
  {$endif}
end;

INLINE (function less_VV (const Val1, Val2: TVectorValue): Boolean);
begin
  {$ifdef USE_GMP}
  if Val1.Alloc < 0 then
    if Val2.Alloc < 0 then
      less_VV := Val1.Val < Val2.Val
    else
      less_VV := mpz_cmp_si (Val2.GMPVal, Val1.Val) > 0
  else
    if Val2.Alloc < 0 then
      less_VV := mpz_cmp_si (Val1.GMPVal, Val2.Val) < 0
    else
      less_VV := mpz_cmp (Val1.GMPVal, Val2.GMPVal) < 0
  {$else}
  less_VV := Val1 < Val2
  {$endif}
end;

{$ifdef USE_GMP}
var
  mpz_tmp, mpz_tmp2: mpz_t;
  V_Tmp1, V_Tmp2: TVectorValue;
{$define Declare_V_Tmp1}
{$define Declare_V_Tmp2}
{$else}
{$define Declare_V_Tmp1 var V_Tmp1: TVectorValue; }
{$define Declare_V_Tmp2 var V_Tmp1, V_Tmp2: TVectorValue; }
{$endif}

INLINE (function eq_VV_neg (const Val1, Val2: TVectorValue): Boolean);
begin
  {$ifdef USE_GMP}
  if Val1.Alloc < 0 then
    if Val2.Alloc < 0 then
      eq_VV_neg := Val1.Val = -Val2.Val
    else
      eq_VV_neg := mpz_cmp_si (Val2.GMPVal, -Val1.Val) = 0
  else
    if Val2.Alloc < 0 then
      eq_VV_neg := mpz_cmp_si (Val1.GMPVal, -Val2.Val) = 0
    else
      begin
        mpz_neg (mpz_tmp, Val2.GMPVal);
        eq_VV_neg := mpz_cmp (Val1.GMPVal, mpz_tmp) = 0
      end
  {$else}
  eq_VV_neg := Val1 = -Val2
  {$endif}
end;

{$ifdef USE_GMP}

INLINE (function mpz_IsZero (const Val: mpz_t): Boolean);
begin
  mpz_IsZero := Val.mp_size = 0
end;

INLINE (function mpz_Negative (const Val: mpz_t): Boolean);
begin
  mpz_Negative := Val.mp_size < 0
end;

INLINE (procedure mpz_add_si (var Dest: mpz_t; const Src1: mpz_t; Src2: NormalInt));
begin
  if Src2 >= 0 then
    mpz_add_ui (Dest, Src1, Src2)
  else
    mpz_sub_ui (Dest, Src1, -Src2)
end;

INLINE (procedure mpz_mul_si (var Dest: mpz_t; const Src1: mpz_t; Src2: NormalInt));
begin
  if Src2 >= 0 then
    mpz_mul_ui (Dest, Src1, Src2)
  else
    begin
      mpz_mul_ui (Dest, Src1, -Src2);
      mpz_neg (Dest, Dest)
    end
end;

INLINE (procedure mpz_tdiv_q_si (var Dest: mpz_t; const Src1: mpz_t; Src2: NormalInt));
begin
  if Src2 >= 0 then
    mpz_tdiv_q_ui (Dest, Src1, Src2)
  else
    begin
      mpz_tdiv_q_ui (Dest, Src1, -Src2);
      mpz_neg (Dest, Dest)
    end
end;

INLINE (procedure mpz_tdiv_qr_si (var DestQ, DestR: mpz_t; const Src1: mpz_t; Src2: NormalInt));
begin
  if Src2 >= 0 then
    mpz_tdiv_qr_ui (DestQ, DestR, Src1, Src2)
  else
    begin
      mpz_tdiv_qr_ui (DestQ, DestR, Src1, -Src2);
      mpz_neg (DestQ, DestQ);
      mpz_neg (DestR, DestR)
    end
end;

INLINE (function mpz_mod_si (const Src1: mpz_t; Src2: NormalInt): MedCard);
begin
  mpz_mod_si := mpz_fdiv_r_ui (mpz_tmp, Src1, Abs (Src2))
end;

{$endif}

INLINE (function IsZero (const Val: TVectorValue): Boolean);
begin
  {$ifdef USE_GMP}
  if Val.Alloc < 0 then
    IsZero := Val.Val = 0
  else
    IsZero := mpz_IsZero (Val.GMPVal)
  {$else}
  IsZero := Val = 0
  {$endif}
end;

INLINE (function Negative (const Val1: TVectorValue): Boolean);
begin
  {$ifdef USE_GMP}
  if Val1.Alloc < 0 then
    Negative := Val1.Val < 0
  else
    Negative := mpz_Negative (Val1.GMPVal)
  {$else}
  Negative := Val1 < 0
  {$endif}
end;

INLINE (procedure Negate (var Dest: TVectorValue; const Src: TVectorValue));
begin
  {$ifdef USE_GMP}
  if Src.Alloc < 0 then
    Set_Int (Dest, -Src.Val)
  else
    begin
      G_Init (Dest);
      mpz_neg (Dest.GMPVal, Src.GMPVal)
    end
  {$else}
  Dest := -Src
  {$endif}
end;

INLINE (procedure Add (var Dest: TVectorValue; const Src1, Src2: TVectorValue));
{$ifdef USE_GMP}
var
  Tmp, Val1, Val2: NormalInt;
  All1, All2: Integer;
begin
  if (Src1.Alloc < 0) and (Src2.Alloc < 0) then
    begin
      Tmp := Src1.Val + Src2.Val;
      if Abs (Tmp) <= MaxVal then
        Set_Int (Dest, Tmp)
      else
        begin
          G_Init (Dest);
          mpz_set_si (Dest.GMPVal, Tmp)
        end
    end
  else
    begin
      Val1 := Src1.Val;
      Val2 := Src2.Val;
      All1 := Src1.Alloc;
      All2 := Src2.Alloc;
      G_Init (Dest);
      if All1 < 0 then
        mpz_add_si (Dest.GMPVal, Src2.GMPVal, Val1)
      else if All2 < 0 then
        mpz_add_si (Dest.GMPVal, Src1.GMPVal, Val2)
      else
        mpz_add (Dest.GMPVal, Src1.GMPVal, Src2.GMPVal)
    end
{$else}
begin
  Dest := Src1 + Src2
{$endif}
end;

INLINE (procedure Add_Check (var Dest: TVectorValue; const Src1, Src2: TVectorValue));
begin
  {$ifdef USE_GMP}
  Add (Dest, Src1, Src2)
  {$else}
  Dest := CheckAdd (Src1, Src2)
  {$endif}
end;

INLINE (procedure Sub (var Dest: TVectorValue; const Src1, Src2: TVectorValue));
{$ifdef USE_GMP}
var
  Tmp, Val1, Val2: NormalInt;
  All1, All2: Integer;
begin
  if (Src1.Alloc < 0) and (Src2.Alloc < 0) then
    begin
      Tmp := Src1.Val - Src2.Val;
      if Abs (Tmp) <= MaxVal then
        Set_Int (Dest, Tmp)
      else
        begin
          G_Init (Dest);
          mpz_set_si (Dest.GMPVal, Tmp)
        end
    end
  else
    begin
      Val1 := Src1.Val;
      Val2 := Src2.Val;
      All1 := Src1.Alloc;
      All2 := Src2.Alloc;
      G_Init (Dest);
      if All1 < 0 then
        begin
          mpz_neg (Dest.GMPVal, Src2.GMPVal);
          mpz_add_si (Dest.GMPVal, Dest.GMPVal, Val1)
        end
      else if All2 < 0 then
        mpz_add_si (Dest.GMPVal, Src1.GMPVal, -Val2)
      else
        mpz_sub (Dest.GMPVal, Src1.GMPVal, Src2.GMPVal)
    end
{$else}
begin
  Dest := Src1 - Src2
{$endif}
end;

INLINE (procedure Mult (var Dest: TVectorValue; const Src1, Src2: TVectorValue));
{$ifdef USE_GMP}
var
  Tmp: LargeInt;
  Val1, Val2: NormalInt;
  All1, All2: Integer;
begin
  if (Src1.Alloc < 0) and (Src2.Alloc < 0) then
    begin
      Tmp := LargeInt (Src1.Val) * LargeInt (Src2.Val);
      if Abs (Tmp) <= MaxVal then
        Set_Int (Dest, Tmp)
      else
        begin
          Val1 := Src1.Val;
          Val2 := Src2.Val;
          G_Init (Dest);
          mpz_set_si (Dest.GMPVal, Val1);
          mpz_mul_si (Dest.GMPVal, Dest.GMPVal, Val2)
        end
    end
  else
    begin
      Val1 := Src1.Val;
      Val2 := Src2.Val;
      All1 := Src1.Alloc;
      All2 := Src2.Alloc;
      G_Init (Dest);
      if All1 < 0 then
        mpz_mul_si (Dest.GMPVal, Src2.GMPVal, Val1)
      else if All2 < 0 then
        mpz_mul_si (Dest.GMPVal, Src1.GMPVal, Val2)
      else
        mpz_mul (Dest.GMPVal, Src1.GMPVal, Src2.GMPVal)
    end
{$else}
begin
  Dest := Src1 * Src2
{$endif}
end;

INLINE (procedure Mult_Check (var Dest: TVectorValue; const Src1, Src2: TVectorValue));
begin
  {$ifdef USE_GMP}
  Mult (Dest, Src1, Src2)
  {$else}
  Dest := CheckMult (Src1, Src2)
  {$endif}
end;

INLINE (procedure Divide (var Dest: TVectorValue; const Src1, Src2: TVectorValue));
{$ifdef USE_GMP}
var
  Val1, Val2: NormalInt;
  All1, All2: Integer;
begin
  if (Src1.Alloc < 0) and (Src2.Alloc < 0) then
    Set_Int (Dest, Src1.Val div Src2.Val)
  else
    begin
      Val1 := Src1.Val;
      Val2 := Src2.Val;
      All1 := Src1.Alloc;
      All2 := Src2.Alloc;
      G_Init (Dest);
      if All1 < 0 then
        begin
          mpz_set_si (Dest.GMPVal, Val1);
          mpz_tdiv_q (Dest.GMPVal, Dest.GMPVal, Src2.GMPVal)
        end
      else if All2 < 0 then
        mpz_tdiv_q_si (Dest.GMPVal, Src1.GMPVal, Val2)
      else
        mpz_tdiv_q (Dest.GMPVal, Src1.GMPVal, Src2.GMPVal);
      G_CheckInt (Dest)
    end
{$else}
begin
  Dest := Src1 div Src2
{$endif}
end;

INLINE (procedure DivExact (var Dest: TVectorValue; const Src1, Src2: TVectorValue));
{$ifdef USE_GMP}
var
  Val1, Val2: NormalInt;
  All1, All2: Integer;
begin
  if (Src1.Alloc < 0) and (Src2.Alloc < 0) then
    Set_Int (Dest, Src1.Val div Src2.Val)
  else
    begin
      Val1 := Src1.Val;
      Val2 := Src2.Val;
      All1 := Src1.Alloc;
      All2 := Src2.Alloc;
      G_Init (Dest);
      if All1 < 0 then
        begin
          mpz_set_si (Dest.GMPVal, Val1);
          mpz_tdiv_q (Dest.GMPVal, Dest.GMPVal, Src2.GMPVal)
        end
      else if All2 < 0 then
        mpz_tdiv_q_si (Dest.GMPVal, Src1.GMPVal, Val2)
      else
        mpz_divexact (Dest.GMPVal, Src1.GMPVal, Src2.GMPVal);
      G_CheckInt (Dest)
    end
{$else}
begin
  Dest := Src1 div Src2
{$endif}
end;

INLINE (function Divisible (const Src1, Src2: TVectorValue): Boolean);
begin
  {$ifdef USE_GMP}
  if (Src1.Alloc < 0) and (Src2.Alloc < 0) then
    Divisible := (Abs (Src2.Val) = 1) or (Src1.Val mod Abs (Src2.Val) = 0)
  else
    begin
      if Src1.Alloc < 0 then
        begin
          mpz_set_si (mpz_tmp, Src1.Val);
          mpz_mod (mpz_tmp, mpz_tmp, Src2.GMPVal);
          Divisible := mpz_IsZero (mpz_tmp)
        end
      else if Src2.Alloc < 0 then
        Divisible := mpz_mod_si (Src1.GMPVal, Src2.Val) = 0
      else
        begin
          mpz_mod (mpz_tmp, Src1.GMPVal, Src2.GMPVal);
          Divisible := mpz_IsZero (mpz_tmp)
        end
    end
  {$else}
  Divisible := (Abs (Src2) = 1) or (Src1 mod Abs (Src2) = 0)
  {$endif}
end;

INLINE (function Divisible_F (var Dest: TVectorValue; const Src1, Src2: TVectorValue): Boolean);
{$ifdef USE_GMP}
var
  Val1, Val2: NormalInt;
  All1, All2: Integer;
begin
  if (Src1.Alloc < 0) and (Src2.Alloc < 0) then
    if (Abs (Src2.Val) = 1) or (Src1.Val mod Abs (Src2.Val) = 0) then
      begin
        Set_Int (Dest, Src1.Val div Src2.Val);
        Divisible_F := True
      end
    else
      Divisible_F := False
  else
    begin
      Val1 := Src1.Val;
      Val2 := Src2.Val;
      All1 := Src1.Alloc;
      All2 := Src2.Alloc;
      G_Init (Dest);
      if All1 < 0 then
        begin
          mpz_set_si (Dest.GMPVal, Val1);
          mpz_tdiv_qr (Dest.GMPVal, mpz_tmp, Dest.GMPVal, Src2.GMPVal)
        end
      else if All2 < 0 then
        mpz_tdiv_qr_si (Dest.GMPVal, mpz_tmp, Src1.GMPVal, Val2)
      else
        mpz_tdiv_qr (Dest.GMPVal, mpz_tmp, Src1.GMPVal, Src2.GMPVal);
      Divisible_F := mpz_IsZero (mpz_tmp);
      G_CheckInt (Dest)
    end
{$else}
begin
  if (Abs (Src2) = 1) or (Src1 mod Abs (Src2) = 0) then
    begin
      Dest := Src1 div Src2;
      Divisible_F := True
    end
  else
    Divisible_F := False
{$endif}
end;

INLINE (procedure ModReduce (var v: TVectorValue; const Modulus: TVectorValue));
begin
  {$ifdef USE_GMP}
  if v.Alloc < 0 then
    begin
      if (Modulus.Alloc < 0) and (Abs (v.Val) >= Abs (Modulus.Val)) then
        v.Val := v.Val mod Abs (Modulus.Val)
    end
  else
    begin
      if Modulus.Alloc < 0 then
        begin
          mpz_set_si (mpz_tmp, Modulus.Val);
          mpz_mod (v.GMPVal, v.GMPVal, mpz_tmp)
        end
      else
        if mpz_sizeinbase (v.GMPVal, 2) >= mpz_sizeinbase (Modulus.GMPVal, 2) then
          mpz_mod (v.GMPVal, v.GMPVal, Modulus.GMPVal);
      G_CheckInt (v)
    end
  {$else}
  if Abs (v) >= Abs (Modulus) then v := v mod Abs (Modulus)
  {$endif}
end;

{$ifndef USE_GMP}
{ Extended Euclidean algorithm
  The return value is the (positive) gcd of a and b.
  c, d, e, f are set such that
  (c d) (a) = (gcd)  and  det (c d) = +/-1
  (e f) (b)   ( 0 )           (e f)        }
INLINE (procedure IntGCDExt (var Dest: TVectorValue; const a0, b0: TVectorValue; var c, d, e, f: TVectorValue));
var a, b, g, h: TVectorValue;
begin
  a := a0;
  b := b0;
  c := 1;
  d := 0;
  e := 0;
  f := 1;
  if a < 0 then
    begin
      a := -a;
      c := -1
    end;
  if b < 0 then
    begin
      b := -b;
      f := -1
    end;
  while b <> 0 do
    begin
      g := a div b;
      h := CheckAdd (c, -CheckMult (g, e));
      c := e;
      e := h;
      h := CheckAdd (d, -CheckMult (g, f));
      d := f;
      f := h;
      h := CheckAdd (a, -CheckMult (g, b));
      a := b;
      b := h
    end;
  Dest := a
end;
{$endif}

INLINE (procedure GCDExt (var Dest: TVectorValue; const Src1, Src2: TVectorValue; var c, d, e, f: TVectorValue));
begin
  {$ifdef USE_GMP}
  if Src1.Alloc < 0 then
    mpz_set_si (mpz_tmp, Src1.Val)
  else
    mpz_set (mpz_tmp, Src1.GMPVal);
  if Src2.Alloc < 0 then
    mpz_set_si (mpz_tmp2, Src2.Val)
  else
    mpz_set (mpz_tmp2, Src2.GMPVal);
  G_Init (Dest);
  G_Init (c);
  G_Init (d);
  G_Init (e);
  G_Init (f);
  mpz_gcdext (Dest.GMPVal, c.GMPVal, d.GMPVal, mpz_tmp, mpz_tmp2);
  mpz_divexact (e.GMPVal, mpz_tmp2, Dest.GMPVal);
  mpz_divexact (f.GMPVal, mpz_tmp, Dest.GMPVal);
  mpz_neg (f.GMPVal, f.GMPVal);
  G_CheckInt (Dest);
  G_CheckInt (c);
  G_CheckInt (d);
  G_CheckInt (e);
  G_CheckInt (f);
  {$else}
  IntGCDExt (Dest, Src1, Src2, c, d, e, f)
  {$endif}
end;

INLINE (procedure GCD (var Dest: TVectorValue; const Src1, Src2: TVectorValue));
{$ifdef USE_GMP}
begin
  if Src1.Alloc < 0 then
    mpz_set_si (mpz_tmp, Src1.Val)
  else
    mpz_set (mpz_tmp, Src1.GMPVal);
  if Src2.Alloc < 0 then
    mpz_set_si (mpz_tmp2, Src2.Val)
  else
    mpz_set (mpz_tmp2, Src2.GMPVal);
  G_Init (Dest);
  mpz_gcd (Dest.GMPVal, mpz_tmp, mpz_tmp2);
  G_CheckInt (Dest)
{$else}
var d1, d2, d3, d4: TVectorValue;
begin
  IntGCDExt (Dest, Src1, Src2, d1, d2, d3, d4)
{$endif}
end;

procedure V_Write (var OutFile: Text; const Val: TVectorValue);
{$ifdef USE_GMP}
var Str: CString;
begin
  if Val.Alloc < 0 then
    Write (OutFile, Val.Val)
  else
    begin
      Str := mpz_get_str (nil, 10, Val.GMPVal);
      {$X+} Write (OutFile, Str); {$X-}
      Dispose (Str)
    end
{$else}
begin
  Write (OutFile, Val)
{$endif}
end;

procedure V_WriteStr (var s: String; const Val: TVectorValue);
{$ifdef USE_GMP}
var Str: CString;
begin
  if Val.Alloc < 0 then
    WriteStr (s, Val.Val)
  else
    begin
      Str := mpz_get_str (nil, 10, Val.GMPVal);
      s := CString2String (Str);
      Dispose (Str)
    end
{$else}
begin
  WriteStr (s, Val)
{$endif}
end;

procedure V_WriteStr2 (var s: String; const Val: TVectorValue);
{$ifdef USE_GMP}
var
  Str: CString;
  Digits: SizeType;
begin
  if Val.Alloc < 0 then
    WriteStr (s, Val.Val)
  else
    begin
      Digits := mpz_sizeinbase (Val.GMPVal, 10);
      if Digits <= 24 then
        begin
          Str := mpz_get_str (nil, 10, Val.GMPVal);
          s := CString2String (Str);
          Dispose (Str)
        end
      else
        WriteStr (s, ' (', Digits, ' digits)')
    end
{$else}
begin
  WriteStr (s, Val)
{$endif}
end;

INLINE (procedure NewSparseVector (var SparseVector: TSparseVector; InitStart: TCount));
begin
  if SparseVector.Size > High (SparseVector.Components^) then Abort ('Vector too long!');
  GetMem (SparseVector.Components, SparseVector.Size * SizeOf (SparseVector.Components^[0]));
  {$ifdef USE_GMP}
  var i: TCount;
  for i := InitStart to SparseVector.Size - 1 do V_Init (SparseVector.Components^[i].Val)
  {$else}
  var Dummy: TCount;
  Dummy := InitStart
  {$endif}
end;

INLINE (procedure CopySparseVector (Source: TSparseVector; var Dest: TSparseVector));
var i: TCount;
begin
  Dest.Count := Source.Count;
  Dest.Size := Dest.Count;
  SetMax (Dest.MaxEl, Source.MaxEl);
  NewSparseVector (Dest, 0);
  for i := 0 to Dest.Count - 1 do
    begin
      Dest.Components^[i].Index := Source.Components^[i].Index;
      Set_V (Dest.Components^[i].Val, Source.Components^[i].Val)
    end
end;

INLINE (procedure ShrinkSparseVector (var Vector: TSparseVector));
begin
  with Vector do
    if Size > Count then
      begin
        {$ifdef USE_GMP}
        var i: Integer;
        for i := Count + 1 to Size - 1 do V_Done (Components^[i].Val);
        {$endif}
        Size := Count;
        ReAllocMem (Components, Size * SizeOf (Components^[0]))
      end
end;

INLINE (procedure GrowSparseVector (var Vector: TSparseVector; NewSize: TCount));
var
  VectorT: TSparseVector;
  i: TCount;
begin
  VectorT.Count := Vector.Count;
  VectorT.Size := Max (NewSize, Vector.Size);
  SetMax (VectorT.MaxEl, Vector.MaxEl);
  NewSparseVector (VectorT, Vector.Size);
  for i := 0 to Vector.Size - 1 do VectorT.Components^[i] := Vector.Components^[i];
  Dispose (Vector.Components);
  Vector := VectorT
end;

INLINE (procedure DisposeSparseVector (SparseVector: TSparseVector; DoneStart: TCount));
begin
  {$ifdef USE_GMP}
  var i: TCount;
  for i := DoneStart to SparseVector.Size - 1 do V_Done (SparseVector.Components^[i].Val);
  {$else}
  var Dummy: TCount;
  Dummy := DoneStart;
  {$endif}
  Dispose (SparseVector.Components)
end;

procedure DisposeMatrix (Matrix: PSparseMatrix);
var Temp: PSparseMatrix;
begin
  while Matrix <> nil do
    begin
      Temp := Matrix;
      Matrix := Matrix^.Next;
      DisposeSparseVector (Temp^.Row, 0);
      Dispose (Temp)
    end
end;

{ Multiply a sparse vector by -1 }
INLINE (procedure SparseNeg (SparseVector: TSparseVector));
var c: TCount;
begin
  for c := 0 to SparseVector.Count - 1 do
    with SparseVector.Components^[c] do Negate (Val, Val)
end;

{ Divide a sparse vector by an integer where it must be known that it is divisible }
INLINE (procedure SparseDivExact (SparseVector: TSparseVector; const Divisor: TVectorValue));
var c: TCount;
begin
  {$ifndef USE_GMP}
  SparseVector.MaxEl := SparseVector.MaxEl div Divisor;
  {$endif}
  for c := 0 to SparseVector.Count - 1 do
    with SparseVector.Components^[c] do DivExact (Val, Val, Divisor)
end;

{$X+}  { Extended syntax for pointer arithmetic }
{ Important: Components that become 0 must not be stored in the result vector }
{$define Store(Idx, NoTest)
  if NoTest or not IsZero (V_Tmp1) then
    begin
      vr^.Index := Idx;
      Set_V (vr^.Val, V_Tmp1);
      Inc (vr)
    end
}

{ Macro for various row operations. Op1 and Op2 are the operations to be
  applied to the elements of SparseVector1 and SparseVector2, respectively,
  which are not combined, and Op12 is the operation that combines elements
  of both vectors. The results of the operations must be stored in V_Tmp1. }
{$define SparseAddGeneric(Declare_V_Tmp, NoTest1, NoTest2, Op1, Op2, Op12)
var
  c1, c2: TCount;
  v1, v2, vr: PComponent;
  Declare_V_Tmp
begin
  v1 := @SparseVector1.Components^[0];
  v2 := @SparseVector2.Components^[0];
  vr := @Res.Components^[0];
  c1 := SparseVector1.Count;
  c2 := SparseVector2.Count;
  if (c1 > 0) and (c2 > 0) then
    repeat
      if v1^.Index = v2^.Index then
        begin
          Op12;
          Store (v1^.Index, False);
          Inc (v1);
          Inc (v2);
          Dec (c1);
          Dec (c2);
          if (c1 = 0) or (c2 = 0) then Break
        end
      else if v1^.Index < v2^.Index then
        begin
          Op1;
          Store (v1^.Index, NoTest1);
          Inc (v1);
          Dec (c1);
          if c1 = 0 then Break
        end
      else
        begin
          Op2;
          Store (v2^.Index, NoTest2);
          Inc (v2);
          Dec (c2);
          if c2 = 0 then Break
        end
    until False;
  while c1 > 0 do
    begin
      Op1;
      Store (v1^.Index, NoTest1);
      Inc (v1);
      Dec (c1)
    end;
  while c2 > 0 do
    begin
      Op2;
      Store (v2^.Index, NoTest2);
      Inc (v2);
      Dec (c2)
    end;
  Res.Count := vr - @Res.Components^[0]
end;
}

{ Add a sparse vector to another }
procedure SparseAdd (const SparseVector1, SparseVector2: TSparseVector;
                     var Res: TSparseVector);
SparseAddGeneric (Declare_V_Tmp1, True, True,
  Set_V (V_Tmp1, v1^.Val),
  Set_V (V_Tmp1, v2^.Val),
  Add (V_Tmp1, v1^.Val, v2^.Val))

{ Add a sparse vector to another modulo Modulus }
procedure SparseAddMod (const SparseVector1, SparseVector2: TSparseVector;
                        var Res: TSparseVector;
                        const Modulus: TVectorValue);
SparseAddGeneric (Declare_V_Tmp1, False, False,
  Set_V (V_Tmp1, v1^.Val),
  Set_V (V_Tmp1, v2^.Val),
  Add (V_Tmp1, v1^.Val, v2^.Val);
  ModReduce (V_Tmp1, Modulus))

{ Subtract a sparse vector from another }
procedure SparseSub (const SparseVector1, SparseVector2: TSparseVector;
                     var Res: TSparseVector);
SparseAddGeneric (Declare_V_Tmp1, True, True,
  Set_V (V_Tmp1, v1^.Val),
  Negate (V_Tmp1, v2^.Val),
  Sub (V_Tmp1, v1^.Val, v2^.Val))

{ Add an integer multiple of a sparse vector to another vector }
procedure SparseAddScaled (const SparseVector1, SparseVector2: TSparseVector;
                           const ScaleFactor2: TVectorValue;
                           var Res: TSparseVector);
SparseAddGeneric (Declare_V_Tmp2, True, True,
  Set_V (V_Tmp1, v1^.Val),
  Mult (V_Tmp1, ScaleFactor2, v2^.Val),
  Mult (V_Tmp2, ScaleFactor2, v2^.Val);
  Add (V_Tmp1, v1^.Val, V_Tmp2))

{ Add an integer multiple of a sparse vector to another vector modulo Modulus }
procedure SparseAddScaledMod (const SparseVector1, SparseVector2: TSparseVector;
                              const ScaleFactor2: TVectorValue;
                              var Res: TSparseVector;
                              const Modulus: TVectorValue);
SparseAddGeneric (Declare_V_Tmp2, False, False,
  Set_V (V_Tmp1, v1^.Val),
  Mult (V_Tmp1, ScaleFactor2, v2^.Val);
  ModReduce (V_Tmp1, Modulus),
  Mult (V_Tmp2, ScaleFactor2, v2^.Val);
  Add (V_Tmp1, v1^.Val, V_Tmp2);
  ModReduce (V_Tmp1, Modulus))

{ Add an integer multiple of a sparse vector to the multiple of another vector }
procedure SparseAddScaledBoth (SparseVector1, SparseVector2: TSparseVector;
                               const ScaleFactor1, ScaleFactor2: TVectorValue;
                               var Res: TSparseVector);
SparseAddGeneric (Declare_V_Tmp2, False, False,
  Mult (V_Tmp1, ScaleFactor1, v1^.Val),
  Mult (V_Tmp1, ScaleFactor2, v2^.Val),
  Mult (V_Tmp1, ScaleFactor1, v1^.Val);
  Mult (V_Tmp2, ScaleFactor2, v2^.Val);
  Add (V_Tmp1, V_Tmp1, V_Tmp2))
{$X-}

{ Add an integer multiple of a matrix column to another column modulo Modulus }
procedure AddColumnScaledMod (Matrix: PSparseMatrix;
                              SourceColumn, DestColumn: TCount;
                              const ScaleFactor: TVectorValue;
                              const Modulus: TVectorValue);
var
  i, j, k: TCount;
  V_Temp: TVectorValue;
begin
  V_Init (V_Temp);
  while Matrix <> nil do
    with Matrix^.Row do
      begin
        i := 0;
        while (i < Count) and (Components^[i].Index < SourceColumn) do Inc (i);
        if (i < Count) and (Components^[i].Index = SourceColumn) then
          begin
            if DestColumn < SourceColumn then j := 0 else j := i;
            while (j < Count) and (Components^[j].Index < DestColumn) do Inc (j);
            if (j < Count) and (Components^[j].Index = DestColumn) then
              with Components^[j] do
                begin
                  Mult_Check (V_Temp, ScaleFactor, Components^[i].Val);
                  Add_Check (Val, Val, V_Temp);
                  ModReduce (Val, Modulus);
                  if IsZero (Val) then
                    begin
                      V_Done (Val);
                      Dec (Count);
                      for k := j to Count - 1 do Components^[k] := Components^[k + 1];
                      V_Init (Components^[Count].Val)
                    end
                end
            else
              begin
                Mult_Check (V_Temp, ScaleFactor, Components^[i].Val);
                ModReduce (V_Temp, Modulus);
                if not IsZero (V_Temp) then
                  begin
                    if Size <= Count then GrowSparseVector (Matrix^.Row, Count + 4);
                    V_Done (Components^[Count].Val);
                    for k := Count - 1 downto j do Components^[k + 1] := Components^[k];
                    Inc (Count);
                    Init_V (Components^[j].Val, V_Temp);
                    Components^[j].Index := DestColumn
                  end
              end
          end;
        Matrix := Matrix^.Next
      end;
  V_Done (V_Temp)
end;

{ Reduce the vector using all rows of the matrix with pivot 1 }
procedure SparseReduce (var Vector: TSparseVector; RemainingMatrix: PSparseMatrix);
var
  Vector2, VectorT: TSparseVector;
  i, NewSize: TCount;
  MaxEl, V_Temp: TVectorValue;
  Val: PVectorValue;
begin
  V_Init (V_Temp);
  SetMax (MaxEl, Vector.MaxEl);
  Vector2.Size := Max (Vector.Size, Size0);
  NewSparseVector (Vector2, 0);
  while RemainingMatrix <> nil do
    begin
      if eq_VI (RemainingMatrix^.Row.Components^[0].Val, 1) then
        begin
          i := 0;
          while (i < Vector.Count) and (Vector.Components^[i].Index < RemainingMatrix^.PivotColumn) do Inc (i);
          if (i < Vector.Count) and (Vector.Components^[i].Index = RemainingMatrix^.PivotColumn) then
            begin
              if Vector2.Size < Vector.Count + RemainingMatrix^.Row.Count then
                begin
                  NewSize := Vector2.Size;
                  repeat
                    NewSize := 2 * NewSize
                  until NewSize >= Vector.Count + RemainingMatrix^.Row.Count;
                  Vector2.Count := 0;
                  GrowSparseVector (Vector2, NewSize)
                end;
              Val := @Vector.Components^[i].Val;
              if eq_VI (Val^, 1) then
                begin
                  CheckMax (MaxEl, @Vector, RemainingMatrix^.Row.MaxEl);
                  SparseSub (Vector, RemainingMatrix^.Row, Vector2)
                end
              else if eq_VI (Val^, -1) then
                begin
                  CheckMax (MaxEl, @Vector, RemainingMatrix^.Row.MaxEl);
                  SparseAdd (Vector, RemainingMatrix^.Row, Vector2)
                end
              else
                begin
                  {$ifndef USE_GMP}
                  CheckMax (MaxEl, @Vector, CheckMult (RemainingMatrix^.Row.MaxEl, Abs (Val^)));
                  {$endif}
                  Negate (V_Temp, Val^);
                  SparseAddScaled (Vector, RemainingMatrix^.Row, V_Temp, Vector2)
                end;
              VectorT := Vector;
              Vector := Vector2;
              Vector2 := VectorT
            end
        end;
      RemainingMatrix := RemainingMatrix^.Next
    end;
  GetMax (Vector);
  DisposeSparseVector (Vector2, 0);
  V_Done (V_Temp)
end;

procedure GetTorsion (Matrix: PSparseMatrix; Rows: TCount; TorsionList: PPTorsionList; Mode: Integer; const Coeff: TVectorValue);
var
  TorsionCount, min_Col, i, j, c, c2: TCount;
  v: PVectorValue;
  min_Row, Temp: PSparseMatrix;
  PTemp: PPSparseMatrix;
  Vector2: TSparseVector;
  v_abs, TorsionVal, det, m_gcd, m_min, m_min_abs, MaxEl, Factor: TVectorValue;
  Msg: TString;
  MsgLevel: Integer;
  IsOne, WasOne: Boolean;

  procedure StoreTorsion;
  begin
    if not eq_VI (TorsionVal, 1) and (TorsionCount <> 0) then
      begin
        New (TorsionList^);
        Init_V (TorsionList^^.TVal, TorsionVal);
        TorsionList^^.Count := TorsionCount;
        TorsionList := @TorsionList^^.Next;
        TorsionList^ := nil
      end
  end;

  { Row := Row - (Row.Components^[Index].Val div m_min) * min_Row^.Row }
  INLINE (procedure AddRow (Index: TCount; var Row: TSparseVector));
  begin
    Vector2.Size := min_Row^.Row.Count + Row.Count;
    NewSparseVector (Vector2, 0);
    Divide (Factor, Row.Components^[Index].Val, m_min);
    Negate (Factor, Factor);
    {$ifndef USE_GMP}
    var MaxEl: TVectorValue;
    SetMax (MaxEl, Row.MaxEl);
    CheckMax (MaxEl, nil, CheckMult (min_Row^.Row.MaxEl, Abs (Factor)));
    {$endif}
    SparseAddScaledMod (Row, min_Row^.Row, Factor, Vector2, det);
    GetMax (Vector2);
    DisposeSparseVector (Row, 0);
    Row := Vector2
  end;

begin
  StartActivity (Commentator, 'torsion', 'Computing torsion coefficients', LvlNormal, Internal_Description);
  i := 0;
  c := 0;
  c2 := 0;
  V_Init (det);
  V_Init (m_min);
  V_Init (m_min_abs);
  V_Init (m_gcd);
  V_Init (Factor);
  V_Init (v_abs);
  V_Init (TorsionVal);
  Set_Int (det, 1);
  Temp := Matrix;
  while Temp <> nil do
    with Temp^ do
      begin
        if (Row.Count = 0) or (Row.Components^[0].Index <> PivotColumn) then Abort ('Internal error #1 in GetTorsion.');
        Mult_Check (det, det, Row.Components^[0].Val);
        Temp := Temp^.Next
      end;
  if IsPrinted (Commentator, LvlUnimp, Internal_Description) <> 0 then
    begin
      V_WriteStr2 (Msg, det);
      ActivityReport (Commentator, 'det (max. minor): ' + Msg, LvlUnimp, PartialResult)
    end;
  Add_Check (det, det, det);
  Set_Int (TorsionVal, 1);
  TorsionCount := 0;
  WasOne := False;
  repeat
    { Find gcd and minimum of the matrix elements }
    with Matrix^.Row.Components^[0] do
      begin
        Set_V (m_min, Val);
        Abs_V (m_min_abs, m_min);
        Set_V (m_gcd, m_min_abs);
        min_Col := Index;
        min_Row := Matrix;
        if IsZero (m_min) then Abort ('Internal error #2 in GetTorsion.')
      end;
    Temp := Matrix;
    while Temp <> nil do
      begin
        for i := 0 to Temp^.Row.Count - 1 do
          with Temp^.Row.Components^[i] do
            begin
              v := @Val;
              Abs_V (v_abs, v^);
              if less_VV (v_abs, m_min_abs) then
                begin
                  Set_V (m_min, v^);
                  Set_V (m_min_abs, v_abs);
                  min_Col := Index;
                  min_Row := Temp;
                  if IsZero (m_min) then Abort ('Internal error #3 in GetTorsion.')
                end;
              if not Divisible (v^, m_gcd) then GCD (m_gcd, m_gcd, v^);
              if eq_VI (v_abs, 1) then Break
            end;
        if eq_VI (m_min_abs, 1) then Break;
        Temp := Temp^.Next
      end;
    { If the gcd is larger than 1, note this and divide all matrix elements by it }
    if not eq_VI (m_gcd, 1) then
      begin
        StoreTorsion;
        Mult (TorsionVal, TorsionVal, m_gcd);
        if (Mode = hmNTorsionRank) and Divisible (TorsionVal, Coeff) then
          begin
            DisposeMatrix (Matrix);
            TorsionCount := Rows - c;
            Break
          end;
        TorsionCount := 0;
        DivExact (det, det, m_gcd);
        DivExact (m_min, m_min, m_gcd);
        Abs_V (m_min_abs, m_min);
        Temp := Matrix;
        while Temp <> nil do
          begin
            SparseDivExact (Temp^.Row, m_gcd);
            Temp := Temp^.Next
          end
      end;
    IsOne := eq_VI (m_min_abs, 1);
    if (MsgStep <> 0) and ((c2 mod (MsgStep div 10) = 0) or (IsOne and not WasOne)) then
      MsgLevel := LvlUnimp
    else
      MsgLevel := LvlBlabla;
    if IsPrinted (Commentator, MsgLevel, Internal_Description) <> 0 then
      begin
        WasOne := IsOne;
        V_WriteStr2 (Msg, m_min_abs);
        ActivityReport (Commentator, 'Current min: ' + Msg, MsgLevel, Internal_Description);
        ProgressReport (Commentator, 'rows', LvlUnimp, c, Rows)
      end;
    Inc (c2);
    if not IsOne then
      begin
        { Find an element in the same column as m_min that is no multiple of m_min }
        Temp := Matrix;
        while Temp <> nil do
          begin
            if Temp <> min_Row then
              begin
                i := 0;
                while (i < Temp^.Row.Count) and (Temp^.Row.Components^[i].Index < min_Col) do Inc (i);
                if i < Temp^.Row.Count then
                  with Temp^.Row.Components^[i] do
                    if (Index = min_Col) and not Divisible (Val, m_min) then Break
              end;
            Temp := Temp^.Next
          end;
        if Temp <> nil then
          begin
            AddRow (i, Temp^.Row);
            Continue
          end;
        { Find an element in the same row as m_min that is no multiple of m_min }
        i := 0;
        while (i < min_Row^.Row.Count) and Divisible (min_Row^.Row.Components^[i].Val, m_min) do Inc (i);
        if i < min_Row^.Row.Count then
          begin
            Divide (Factor, min_Row^.Row.Components^[i].Val, m_min);
            Negate (Factor, Factor);
            AddColumnScaledMod (Matrix, min_Col, min_Row^.Row.Components^[i].Index, Factor, det);
            Continue
          end;
        { Find an element in another column and row that is no multiple of m_min }
        Temp := Matrix;
        while Temp <> nil do
          begin
            i := 0;
            while (i < Temp^.Row.Count) and Divisible (Temp^.Row.Components^[i].Val, m_min) do Inc (i);
            if i < Temp^.Row.Count then Break;
            Temp := Temp^.Next
          end;
        if Temp = nil then Abort ('Internal error #4 in GetTorsion.');
        j := 0;
        while (j < Temp^.Row.Count) and (Temp^.Row.Components^[j].Index < min_Col) do Inc (j);
        if (j < Temp^.Row.Count) and (Temp^.Row.Components^[j].Index = min_Col) then
          AddRow (j, Temp^.Row);
        Vector2.Size := min_Row^.Row.Count + Temp^.Row.Count;
        NewSparseVector (Vector2, 0);
        SetMax (MaxEl, min_Row^.Row.MaxEl);
        CheckMax (MaxEl, nil, Temp^.Row.MaxEl);
        SparseAddMod (min_Row^.Row, Temp^.Row, Vector2, det);
        GetMax (Vector2);
        DisposeSparseVector (min_Row^.Row, 0);
        min_Row^.Row := Vector2
        { Now there is an element in the same row as m_min that is no
          multiple of m_min. In the next iteration m_min can be reduced. }
      end
    else
      begin
        { Now m_min = +/-1, i.e. it divides all matrix elements.
          One column and one row can be removed. }
        PTemp := @Matrix;
        while PTemp^ <> nil do
          if PTemp^ = min_Row then
            PTemp^ := PTemp^^.Next
          else
            begin
              i := 0;
              while (i < PTemp^^.Row.Count) and (PTemp^^.Row.Components^[i].Index <min_Col) do Inc (i);
              if (i < PTemp^^.Row.Count) and (PTemp^^.Row.Components^[i].Index = min_Col) then
                AddRow (i, PTemp^^.Row);
              PTemp := @PTemp^^.Next
            end;
        DisposeSparseVector (min_Row^.Row, 0);
        Dispose (min_Row);
        Inc (TorsionCount);
        Inc (c)
      end
  until Matrix = nil;
  StoreTorsion;
  V_Done (det);
  V_Done (m_min);
  V_Done (m_min_abs);
  V_Done (m_gcd);
  V_Done (Factor);
  V_Done (v_abs);
  V_Done (TorsionVal);
  StopActivity (Commentator, 'Torsion coefficients computed', LvlNormal, Internal_Description)
end;

function _(GetNextRow) (const BigFaces, SmallFaces: TFaces; var NextRow: TSparseVector; MinSize: TCount; var RowCount, RowDoneCount: TCount; procedure StatusProc): Boolean;
var
  Temp: PSparseIntVector;
  i, c: TCount;
begin
  repeat
    if RowCount >= __(FacesCount) (BigFaces) then
      begin
        _(GetNextRow) := False;
        Exit
      end;
    Temp := __(GetOneRow) (BigFaces, SmallFaces, RowCount);
    Inc (RowCount);
    c := Temp^.Count;
    if c = 0 then
      begin
        Inc (RowDoneCount);
        if (MsgStep <> 0) and (RowDoneCount mod MsgStep = 0) then StatusProc;
        Dispose (Temp)
      end
  until c <> 0;
  NextRow.Count := c;
  NextRow.Size := Max (NextRow.Count, MinSize);
  NewSparseVector (NextRow, 0);
  for i := 0 to c - 1 do
    begin
      NextRow.Components^[i].Index := Temp^.Elements[i].Index;
      Set_Int (NextRow.Components^[i].Val, Temp^.Elements[i].Val)
    end;
  GetMax (NextRow);
  Dispose (Temp);
  _(GetNextRow) := True
end;

{ Compute the Smith normal form of a sparse matrix }
procedure _(SmithNormalForm) (const BigFaces, SmallFaces: TFaces; MaxRank, Dim: TCount;
                              Mode: Integer; const Coeff: TVectorValue;
                              var PivotCount: TCount; var TorsionList: PTorsionList);
var
  PivotNot1Count, i, LastCol, RowCount, LastRowDoneCount, RowDoneCount: TCount;
  ReducedMatrixStartRow: PPSparseMatrix;
  Temp, ReducedMatrix, DeferredVectors, DeferredVectors1, TorsionMatrix: PSparseMatrix;
  Vector, Vector2: TSparseVector;
  pg, vg, p0, v0, Factor, V_Temp: TVectorValue;
  Msg: TString;
  Pass: Integer;

  procedure RowStatus;
  var Msg, Msg2: TString;
  begin
    { When rows are deferred, we don't want multiple messages for the same number of rows done }
    if RowDoneCount = LastRowDoneCount then Exit;
    LastRowDoneCount := RowDoneCount;
    if Pass <= 2 then
      Msg2 := ''
    else
      WriteStr (Msg2, ', rk: ', PivotCount);
    WriteStr (Msg, 'pass ', Pass, ': torsion-free: ', PivotCount - PivotNot1Count,
                   Msg2, ', max.: ', MaxRank);
    ActivityReport (Commentator, Msg, LvlUnimp, Internal_Description);
    ProgressReport (Commentator, 'rows', LvlUnimp, RowDoneCount, __(FacesCount) (BigFaces))
  end;

  procedure DoOneRow (Vector: TSparseVector);
  var
    Index1, j, NewSize: TCount;
    NewRow: PSparseMatrix;
    ReducedMatrixRow: PPSparseMatrix;
    VectorT: TSparseVector;
    Val, PivotValue: PVectorValue;
    MaxEl: TVectorValue;
    m: Integer;
    Defer: Boolean;
  begin
    if Pass <= 2 then
      m := MsgStep
    else
      m := MsgStep div 100;
    if (MsgStep <> 0) and (RowDoneCount mod m = 0) then RowStatus;
    SetMax (MaxEl, Vector.MaxEl);
    Index1 := Vector.Components^[0].Index;
    if Index1 < LastCol then ReducedMatrixStartRow := @ReducedMatrix;
    while (ReducedMatrixStartRow^ <> nil) and (Index1 > ReducedMatrixStartRow^^.PivotColumn) do
      ReducedMatrixStartRow := @ReducedMatrixStartRow^^.Next;
    LastCol := Index1;
    ReducedMatrixRow := ReducedMatrixStartRow;
    Defer := False;
    while ReducedMatrixRow^ <> nil do
      if Index1 > ReducedMatrixRow^^.PivotColumn then
        ReducedMatrixRow := @ReducedMatrixRow^^.Next
      else if Index1 < ReducedMatrixRow^^.PivotColumn then
        Break
      else if Pass = 1 then
        begin
          Defer := True;
          Break
        end
      else
        begin
          if Vector2.Size < Vector.Count + ReducedMatrixRow^^.Row.Count then
            begin
              NewSize := Vector2.Size;
              repeat
                NewSize := 2 * NewSize
              until NewSize >= Vector.Count + ReducedMatrixRow^^.Row.Count;
              Vector2.Count := 0;
              GrowSparseVector (Vector2, NewSize)
            end;
          Val := @Vector.Components^[0].Val;
          PivotValue := @ReducedMatrixRow^^.Row.Components^[0].Val;
          if eq_VV (Val^, PivotValue^) then
            begin
              CheckMax (MaxEl, @Vector, ReducedMatrixRow^^.Row.MaxEl);
              SparseSub (Vector, ReducedMatrixRow^^.Row, Vector2)
            end
          else if eq_VV_neg (Val^, PivotValue^) then
            begin
              CheckMax (MaxEl, @Vector, ReducedMatrixRow^^.Row.MaxEl);
              SparseAdd (Vector, ReducedMatrixRow^^.Row, Vector2)
            end
          else if Divisible_F (Factor, Val^, PivotValue^) then
            begin
              Negate (Factor, Factor);
              {$ifndef USE_GMP}
              CheckMax (MaxEl, @Vector, CheckMult (ReducedMatrixRow^^.Row.MaxEl, Abs (Factor)));
              {$endif}
              SparseAddScaled (Vector, ReducedMatrixRow^^.Row, Factor, Vector2)
            end
          else if Divisible_F (Factor, PivotValue^, Val^) then
            begin
              if Negative (Val^) then
                SparseNeg (Vector)
              else
                Negate (Factor, Factor);
              if eq_VI (Val^, 1) then Dec (PivotNot1Count);
              VectorT := ReducedMatrixRow^^.Row;
              CopySparseVector (Vector, ReducedMatrixRow^^.Row);
              if Vector.Size < VectorT.Size then
                begin
                  DisposeSparseVector (Vector, 0);
                  Vector := VectorT
                end
              else
                begin
                  Vector.Count := VectorT.Count;
                  for j := 0 to Vector.Count - 1 do
                    begin
                      V_Done (Vector.Components^[j].Val);
                      Vector.Components^[j] := VectorT.Components^[j]
                    end;
                  DisposeSparseVector (VectorT, Vector.Count)
                end;
              {$ifndef USE_GMP}
              GetMax (Vector);
              SetMax (MaxEl, Vector.MaxEl);
              CheckMax (MaxEl, nil, CheckMult (ReducedMatrixRow^^.Row.MaxEl, Abs (Factor)));
              {$endif}
              SparseAddScaled (Vector, ReducedMatrixRow^^.Row, Factor, Vector2)
            end
          else
            begin
              GCDExt (V_Temp, PivotValue^, Val^, pg, vg, p0, v0);
              if eq_VI (V_Temp, 1) then Dec (PivotNot1Count);
              {$ifndef USE_GMP}
              GetMax (Vector);
              SetMax (MaxEl, CheckMult (Vector.MaxEl, Abs (vg)));
              CheckMax (MaxEl, nil, CheckMult (ReducedMatrixRow^^.Row.MaxEl, Abs (pg)));
              {$endif}
              SparseAddScaledBoth (ReducedMatrixRow^^.Row, Vector, pg, vg, Vector2);
              CopySparseVector (Vector2, VectorT);
              {$ifndef USE_GMP}
              SetMax (MaxEl, CheckMult (Vector.MaxEl, Abs (v0)));
              CheckMax (MaxEl, nil, CheckMult (ReducedMatrixRow^^.Row.MaxEl, Abs (p0)));
              {$endif}
              SparseAddScaledBoth (ReducedMatrixRow^^.Row, Vector, p0, v0, Vector2);
              DisposeSparseVector (ReducedMatrixRow^^.Row, 0);
              ReducedMatrixRow^^.Row := VectorT
            end;
          VectorT := Vector;
          Vector := Vector2;
          Vector2 := VectorT;
          if Vector.Count = 0 then Break;
          if Index1 >= Vector.Components^[0].Index then Abort ('Internal error in DoOneRow.');
          Index1 := Vector.Components^[0].Index;
          ReducedMatrixRow := @ReducedMatrixRow^^.Next
        end;
    if Vector.Count <> 0 then
      begin
        GetMax (Vector);
        Val := @Vector.Components^[0].Val;
        if Negative (Val^) then SparseNeg (Vector);
        New (NewRow);
        if Defer or not eq_VI (Val^, 1) then
          if Defer or (Pass <= 2) then
            begin
              ShrinkSparseVector (Vector);
              NewRow^.Next := DeferredVectors;
              NewRow^.Row := Vector;
              DeferredVectors := NewRow;
              Exit
            end
          else
            Inc (PivotNot1Count);
        with NewRow^ do
          begin
            Next := ReducedMatrixRow^;
            CopySparseVector (Vector, Row);
            PivotColumn := Index1
          end;
        ReducedMatrixRow^ := NewRow;
        Inc (PivotCount)
      end;
    DisposeSparseVector (Vector, 0);
    Inc (RowDoneCount)
  end;

begin
  TorsionList := nil;
  RowCount := 0;
  RowDoneCount := 0;
  LastRowDoneCount := -1;
  WriteStr (Msg, 'Finding rank of boundary map d', Dim);
  ActivityReport (Commentator, Msg, LvlNormal, Internal_Description);
  StartActivity (Commentator, 'elimination', 'Triangulating matrix by modified Gauss elimination', LvlNormal, Internal_Description);
  ReducedMatrix := nil;
  Vector2.Size := Size0;
  NewSparseVector (Vector2, 0);
  PivotCount := 0;
  PivotNot1Count := 0;
  ReducedMatrixStartRow := @ReducedMatrix;
  LastCol := 0;
  V_Init (Factor);
  V_Init (pg);
  V_Init (vg);
  V_Init (p0);
  V_Init (v0);
  V_Init (V_Temp);
  DeferredVectors := nil;
  Pass := 1;
  while (PivotCount - PivotNot1Count < MaxRank)
        and _(GetNextRow) (BigFaces, SmallFaces, Vector, Size0, RowCount, RowDoneCount, RowStatus) do
    DoOneRow (Vector);
  for Pass := 2 to 3 do
    begin
      DeferredVectors1 := DeferredVectors;
      DeferredVectors := nil;
      while DeferredVectors1 <> nil do
        begin
          Temp := DeferredVectors1;
          DeferredVectors1 := DeferredVectors1^.Next;
          if PivotCount - PivotNot1Count < MaxRank then
            DoOneRow (Temp^.Row)
          else
            DisposeSparseVector (Temp^.Row, 0);
          Dispose (Temp)
        end;
      RowStatus
    end;
  if DeferredVectors <> nil then Abort ('Internal error in SmithNormalForm.');
  V_Done (Factor);
  V_Done (pg);
  V_Done (vg);
  V_Done (p0);
  V_Done (v0);
  V_Done (V_Temp);
  DisposeSparseVector (Vector2, 0);
  StopActivity (Commentator, 'Matrix triangulated', LvlNormal, Internal_Description);
  WriteStr (Msg, 'Rank d', Dim, ' = ', PivotCount);
  ActivityReport (Commentator, Msg, LvlImp, PartialResult);
  if (PivotNot1Count = 0) or (Mode = hmBetti) then
    DisposeMatrix (ReducedMatrix)
  else
    begin
      WriteStr (Msg, 'Searching for torsion of rank <= ', PivotNot1Count);
      ActivityReport (Commentator, Msg, LvlImp, Internal_Description);
      StartActivity (Commentator, 'reduce', 'Reducing matrix', LvlNormal, Internal_Description);
      i := 0;
      TorsionMatrix := nil;
      while ReducedMatrix <> nil do
        begin
          Temp := ReducedMatrix;
          ReducedMatrix := ReducedMatrix^.Next;
          if eq_VI (Temp^.Row.Components^[0].Val, 1) then
            begin
              DisposeSparseVector (Temp^.Row, 0);
              Dispose (Temp)
            end
          else
            begin
              SparseReduce (Temp^.Row, ReducedMatrix);
              Temp^.Next := TorsionMatrix;
              TorsionMatrix := Temp;
              Inc (i);
              ProgressReport (Commentator, 'rows', LvlUnimp + 1, i, PivotNot1Count)
            end
        end;
      StopActivity (Commentator, 'Matrix reduced', LvlNormal, Internal_Description);
      GetTorsion (TorsionMatrix, PivotNot1Count, @TorsionList, Mode, Coeff)
    end
end;

procedure _(WriteStrTorsion) (var s: String; TorsionList: PTorsionList; Flag: Boolean);
var Temp: TString;
begin
  s := '';
  while TorsionList <> nil do
    begin
      if Flag then s := s + ' + ';
      Flag := True;
      s := s + 'Z_';
      V_WriteStr (Temp, TorsionList^.TVal);
      s := s + Temp;
      if TorsionList^.Count <> 1 then
        begin
          WriteStr (Temp, s, '^', TorsionList^.Count);
          s := Temp
        end;
      TorsionList := TorsionList^.Next
    end
end;

procedure _(WriteHomology) (var OutFile: Text; const Homology: _G_(THomTable); MinDim, MaxDim: TCount);

  procedure WriteTorsionList (var OutFile: Text; TorsionList: PTorsionList);
  begin
    while TorsionList <> nil do
      begin
        Write (OutFile, ', [');
        V_Write (OutFile, TorsionList^.TVal);
        Write (OutFile, ', ', TorsionList^.Count, ']');
        TorsionList := TorsionList^.Next
      end
  end;

var j: Dim1;
begin
  Write (OutFile, '[');
  for j := MinDim to MaxDim do
    begin
      if j <> MinDim then Write (OutFile, ', ');
      with Homology[j] do
        begin
          Write (OutFile, '[', Betti);
          WriteTorsionList (OutFile, Torsion);
          Write (OutFile, ']')
        end
    end;
  Write (OutFile, ']')
end;

procedure WriteStrTorsionList2 (var s: String; TorsionList: PTorsionList);
var s1, s2: TString;
begin
  s := '';
  while TorsionList <> nil do
    begin
      V_WriteStr (s1, TorsionList^.TVal);
      if TorsionList^.Count <> 1 then
        WriteStr (s2, '^', TorsionList^.Count)
      else
        s2 := '';
      s := s + ' + Z_' + s1 + s2;
      TorsionList := TorsionList^.Next
    end
end;

procedure WriteStrHomology2 (var s: String; const Homology: _G_(THomTable); MinDim, MaxDim: TCount);
var
  j: Dim1;
  s1, s2: TString;
begin
  s := '[';
  for j := MinDim to MaxDim do
    begin
      if j <> MinDim then s := s + ', ';
      with Homology[j] do
        if Betti < 0 then
          s := s + '?'
        else
          begin
            WriteStrTorsionList2 (s1, Torsion);
            WriteStr (s2, Betti);
            s := s + '[' + s2 + s1 + ']'
          end
    end;
  s := s + ']'
end;

function _(HomologyCompute) (MinDim, MaxDim: TCount; const Facets: TFacets; HomReduced: Integer) = Res: TString;
var
  FaceNumbers, Ranks: array [Dim0] of TCount;
  BigFaces, SmallFaces: TFaces;
  MaxCard, i: Dim0;
  c: Integer;
  MaxD, MinD, MaxRank: TCount;
  Homology: _G_(THomTable);
  STATIC (DummyCoeff, TVectorValue);

  function GetBetti (i: Dim0): TCount;
  var Tmp: TCount;
  begin
    if (i < 1) or (i > MaxCard) then
      GetBetti := FaceNumbers[i]
    else if (FaceNumbers[i] < 0) or (Ranks[i] < 0) or (Ranks[i - 1] < 0) then
      GetBetti := -1
    else
      begin
        Tmp := FaceNumbers[i] - Ranks[i] - Ranks[i - 1];
        if Tmp < 0 then
          Abort ('Internal error: ranks bigger than possible!');
        GetBetti := Tmp
      end
  end;

  procedure WriteH (i: Dim0);
  var
    Tmp: TCount;
    Msg, Msg2: TString;
  begin
    Tmp := GetBetti (i);
    if Tmp >= 0 then
      begin
        WriteStr (Msg, 'H_', i - 1, ' = ', Tmp);
        WriteStrTorsionList2 (Msg2, Homology[i].Torsion);
        ActivityReport (Commentator, Msg + Msg2 + '.', LvlAlways, PartialResult)
      end
  end;

begin
  for c := 0 to MaximumDim do
    begin
      FaceNumbers[c] := -1;
      Ranks[c] := -1
    end;
  Ranks[0] := HomReduced;
  MaxCard := __(GetDimension) (Facets) + 1;
  if (MaxDim = 0) or (MaxDim >= MaxCard) then MaxDim := MaxCard - 1;
  MinD := Min (MaxCard,Max (1, MinDim));
  MaxD := Min (MaxCard,Max (0, MaxDim + 2));
  for i := 0 to MaximumDim do Homology[i].Torsion := nil;
  __(GetFaces) (Facets, BigFaces, MinD - 1);
  FaceNumbers[MinD] := __(FacesCount) (BigFaces);
  if MaxCard >= 2 then
    for i := MinD to MaxD - 1 do
      begin
        SmallFaces := BigFaces;
        __(GetFaces) (Facets, BigFaces, i);
        FaceNumbers[i + 1] := __(FacesCount) (BigFaces);
        MaxRank := FaceNumbers[i] - Max (Ranks[i - 1], 0);
        _(SmithNormalForm) (BigFaces, SmallFaces, MaxRank, i, hmGroup, DummyCoeff,
                            Ranks[i], Homology[i].Torsion);
        WriteH (i);
        __(DisposeFaces) (SmallFaces)
      end;
  __(DisposeFaces) (BigFaces);
  for i := MaxCard + 1 to MaximumDim do FaceNumbers[i] := 0;
  Ranks[MaxCard] := 0;
  WriteH (MaxD);
  for i := 0 to MaximumDim do Homology[i].Betti := GetBetti (i);
  WriteStrHomology2 (Res, Homology, 1, MaxCard)
end;

{$ifdef USE_GMP}

to begin do
  begin
    mpz_init (mpz_tmp);
    mpz_init (mpz_tmp2);
    V_Init (V_Tmp1);
    V_Init (V_Tmp2)
  end;

to end do
  begin
    mpz_clear (mpz_tmp);
    mpz_clear (mpz_tmp2);
    V_Done (V_Tmp1);
    V_Done (V_Tmp2)
  end;

{$endif}

end.
