{ Copyright 1997-2003 by Frank Heckenbach <heckenb@mi.uni-erlangen.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 2.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. }

{$gnu-pascal,I+,W-,R-,W+}  { No automatic range checking, ranges are checked explicitly here. }

unit Simplicial;

interface

uses GPC, Tools, Commentator_Pas;

type
  PPFaceArray = ^TPFaceArray;
  TPFaceArray = array [0 .. MaxVarSize div SizeOf (PFace) - 1] of PFace;

  { Type for a set of simplices, stored as an ordered dynamic array
    with a field containing their number }
  TFaces = record
    Count: TCount;
    FaceP: PPFaceArray
  end;
  SimplicialTFaces = TFaces;

  TFacets = TFaces;

{ Return the cardinality of a set of faces }
function FacesCount (const Faces: TFaces): TCount;

{ Return the dimension of a complex }
function GetDimension (const Facets: TFacets): Dim0;

{ Generate all Dim dimensional simplices of the complex with the given facets }
procedure GetFaces (const Facets: TFacets; var Faces: TFaces; Dim: Dim0);

{ Return one row of the boundary matrix }
function GetOneRow (const BigFaces, SmallFaces: TFaces; RowCount: TCount): PSparseIntVector;

{ Dispose of a set of simplices }
procedure DisposeFaces (const Faces: TFaces);

{ Read a complex from the input }
procedure ReadComplex (var Facets: TFaces);

implementation

{ Create a new set of simplices of the given size }
procedure NewFaces (var Faces: TFaces; Count: TCount);
begin
  if Count > High (Faces.FaceP^) then Abort ('Too many faces!');
  Faces.Count := Count;
  GetMem (Faces.FaceP, Faces.Count * SizeOf (Faces.FaceP^[0]))
end;

procedure DisposeFaces (const Faces: TFaces);
var i: TCount;
begin
  for i := 0 to Faces.Count - 1 do Dispose (Faces.FaceP^[i]);
  Dispose (Faces.FaceP)
end;

{ Return the cardinality of a face }
function FaceCount (const Face: TFace) = Res: n0;
begin
  Res := 0;
  while (Res < MaxN) and (Face[Res] <> 0) do Inc (Res)
end;

function FacesCount (const Faces: TFaces): TCount;
begin
  FacesCount := Faces.Count
end;

procedure FaceCopy (const Src: TFace; var Dest: TFace);
var i: n0;
begin
  for i := 0 to FaceCount (Src) do Dest[i] := Src[i]
end;

{ Return the pointer to a simplex from a set }
function FaceGet (const Faces: TFaces; Index: TCount): PFace;
begin
  FaceGet := Faces.FaceP^[Index]
end;

{ Return a pointer to the pointer to a simplex from a set }
function FacePGet (const Faces: TFaces; Index: TCount): PPFace;
begin
  FacePGet := @Faces.FaceP^[Index]
end;

{ Compare two simplices lexicographically }
function FaceLess (const a, b: TFace): Boolean;
var i: n0;
begin
  i := 0;
  while (a[i] <> 0) and (a[i] = b[i]) do Inc (i);
  FaceLess := a[i] < b[i]
end;

{ Test whether to faces are identical }
function FaceEqual (const a, b: TFace): Boolean;
var i: n0;
begin
  i := 0;
  while (a[i] <> 0) and (a[i] = b[i]) do Inc (i);
  FaceEqual := (a[i] = 0) and (b[i] = 0)
end;

function GetDimension (const Facets: TFacets): Dim0;
var
  d: Dim0;
  c: TCount;
  Msg: TString;
begin
  d := 0;
  for c := 0 to Facets.Count - 1 do
    d := Max (d, FaceCount (FaceGet (Facets, c)^));
  WriteStr (Msg, 'Simplicial complex of dimension ', d - 1, ' with ', Facets.Count, ' facets');
  ActivityReport (Commentator, Msg, LvlNormal, Internal_Description);
  GetDimension := d - 1
end;

procedure GetFaces (const Facets: TFacets; var Faces: TFaces; Dim: Dim0);
const
  MaxI = 2 * MaxN;
type
  { Type for a list of simplices }
  PPFaceList = ^PFaceList;
  PFaceList = ^TFaceList;
  TFaceList = record
    Next: PFaceList;
    FaceP: PFace
  end;

  { Type for a list of simplicies and their number }
  TFaceMerge = record
    Count: TCount;
    List: PFaceList
  end;

var
  c: TCount;
  FacetCard: n0;
  j, k: 0 .. MaxN + 1;
  i: 1 .. MaxI;
  si: n1;
  Facet, SubSet: TFace;
  SubSetIndex: array [n0] of n0;
  Msg, FacesStr: TString;

  { An array of lists of simplices. This array is used in such a way
    that for each i the i-th list never contains more than 2^(i-1)
    elements. New elements are always inserted into the first list.
    Whenever a list grows too big, it is merged with the following
    list and sorted. In this process, when elements exist in both
    lists, one copy of them is removed. This process is repeated
    with the following lists until each list's size is below its
    limit. Many simplices, especially in lower and medium
    dimensions, are subsets of many facets, so they are generated a
    number of times by the main procedure. Therefore, first
    generating all subsets of the given dimension of all facets and
    afterwards removing duplicates would take much more memory. }
  FM: array [1 .. MaxI] of TFaceMerge;

  { Merge and sort two lists.
    The return value is the number of items removed. }
  function Merge (Dest: PPFaceList; Source: PFaceList): TCount;
  var
    t1, t2: PFaceList;
    c: TCount;
  begin
    c := 0;
    while (Source <> nil) and (Dest^ <> nil) do
      if FaceEqual (Source^.FaceP^, Dest^^.FaceP^) then
        begin
          t1 := Source;
          Source := Source^.Next;
          Dispose (t1^.FaceP);
          Dispose (t1);
          Inc (c)
        end
      else
        begin
          if FaceLess (Source^.FaceP^, Dest^^.FaceP^) then
            begin
              t1 := Dest^;
              t2 := Source^.Next;
              Dest^ := Source;
              Source^.Next := t1;
              Source := t2
            end;
          Dest := @Dest^^.Next
        end;
    if Source <> nil then Dest^ := Source;
    Merge := c
  end;

  { Insert a new simplex and merge lists when necessary }
  procedure NewFace (const NFace: TFace);
  var
    pn: PFaceList;
    i: 1 .. MaxI;
    c, j: n0;
  begin
    if (FM[1].Count <> 0) and FaceEqual (NFace, FM[1].List^.FaceP^) then Exit;
    New (pn);
    c := FaceCount (NFace);
    GetMem (pn^.FaceP, (c + 1) * SizeOf (pn^.FaceP^[0]));
    for j := 0 to c do pn^.FaceP^[j] := NFace[j];
    pn^.Next := nil;
    Inc (FM[1].Count);
    if FM[1].Count = 1 then
      FM[1].List := pn
    else
      begin
        if FaceLess (NFace, FM[1].List^.FaceP^) then
          begin
            pn^.Next := FM[1].List;
            FM[1].List := pn
          end
        else
          FM[1].List^.Next := pn;
        i := 1;
        while (i < MaxI) and (FM[i].Count > TCount (1) shl (i - 1)) do
          begin
            if FM[i + 1].Count = 0 then
              FM[i + 1] := FM[i]
            else
              Inc (FM[i + 1].Count, FM[i].Count - Merge (@FM[i + 1].List, FM[i].List));
            FM[i].Count := 0;
            FM[i].List := nil;
            Inc (i)
          end
      end
  end;

  { Merge all lists }
  procedure CollectFaces;
  var
    i, j: 1 .. MaxI + 1;
    p, pt: PFaceList;
    c: TCount;
  begin
    i := 1;
    while (i < MaxI) and (FM[i].Count = 0) do Inc (i);
    c := FM[i].Count;
    j := i;
    while i < MaxI do
      begin
        Inc (i);
        while (i <= MaxI) and (FM[i].Count = 0) do Inc (i);
        if i <= MaxI then
          begin
            Inc (c, FM[i].Count - Merge (@FM[i].List, FM[j].List));
            j := i
          end
      end;
    p := FM[j].List;
    NewFaces (Faces, c);
    c := 0;
    while p <> nil do
      begin
        FacePGet (Faces, c)^ := p^.FaceP;
        Inc (c);
        pt := p;
        p := p^.Next;
        Dispose (pt)
      end
  end;

  procedure DoReport;
  begin
    ProgressReport (Commentator, 'facets', LvlUnimp, c + 1, Facets.Count)
  end;

begin
  WriteStr (FacesStr, 'faces of dimension ', Dim);
  StartActivity (Commentator, 'faces', 'Finding ' + FacesStr, LvlNormal, Internal_Description);
  for i := 1 to MaxI do
    with FM[i] do
      begin
        Count := 0;
        List := nil
      end;
  for j := 0 to MaxN do SubSet[j] := 0;
  for c := 0 to Facets.Count - 1 do
    begin
      FaceCopy (FaceGet (Facets, c)^, Facet);
      FacetCard := FaceCount (Facet);
      if FacetCard > Dim then
        begin
          for j := 0 to Dim do
            begin
              SubSetIndex[j] := j;
              SubSet[j] := Facet[j]
            end;
          repeat
            NewFace (SubSet);
            j := Dim + 1;
            while (j <> 0) and (SubSetIndex[j - 1] = FacetCard - (Dim + 1) + j - 1) do Dec (j);
            if j <> 0 then
              begin
                Inc (SubSetIndex[j - 1]);
                si := SubSetIndex[j - 1];
                SubSet[j - 1] := Facet[si];
                for k := j to Dim do
                  begin
                    Inc (si);
                    SubSetIndex[k] := si;
                    SubSet[k] := Facet[si]
                  end
              end
          until j = 0
        end;
      if (MsgStep <> 0) and ((c + 1) mod MsgStep = 0) then DoReport
    end;
  if (MsgStep <> 0) and (Facets.Count mod MsgStep <> 0) then DoReport;
  CollectFaces;
  WriteStr (Msg, 'Found ', Faces.Count, ' ', FacesStr);
  StopActivity (Commentator, Msg, LvlNormal, Internal_Description)
end;

{ Search a simplex in a set of simplices.
  Binary search since this function is time-critical. }
function FindFace (const Face: TFace; const Faces: TFaces): TCount;
var
  a, b, c: TCount;
  p: PFace;
begin
  a := 0;
  b := Faces.Count - 1;
  repeat
    c := (a + b) div 2;
    p := FaceGet (Faces, c);
    if FaceEqual (Face, p^) then Return c;
    if FaceLess (Face, p^) then
      b := c - 1
    else
      a := c + 1
  until a > b;
  Abort ('Internal error in FindFace.');
  FindFace := 0
end;

function GetOneRow (const BigFaces, SmallFaces: TFaces; RowCount: TCount) = Res: PSparseIntVector;
var
  i, k, s, Temp: TCount;
  Sign: -1 .. 1;
  SmallFace: TFace;
begin
  FaceCopy (FaceGet (BigFaces, BigFaces.Count - 1 - RowCount)^, SmallFace);
  New (Res, FaceCount (SmallFace));
  Res^.Count := Res^.Size;
  k := 0;
  Sign := 1;
  for i := Res^.Count downto 1 do
    begin
      Temp := SmallFace[i - 1];
      SmallFace[i - 1] := k;
      k := Temp;
      s := FindFace (SmallFace, SmallFaces);
      if (i <> Res^.Count) and (Res^.Elements[Res^.Count - i - 1].Index >= s) then
        Abort ('Internal error with vertex ordering.');
      with Res^.Elements[Res^.Count - i] do
        begin
          Index := s;
          Val := Sign
        end;
      Sign := -Sign
    end
end;

procedure ReadComplex (var Facets: TFaces);
var
  FacetList, pf: PFacetList;
  FacetCount, PointCount, c: TCount;
begin
  FacetList := ParseFacets ('facet', True, FacetCount, PointCount);
  NewFaces (Facets, FacetCount);
  c := 0;
  while FacetList <> nil do
    begin
      FacePGet (Facets, c)^ := FacetList^.Facet;
      Inc (c);
      pf := FacetList;
      FacetList := FacetList^.Next;
      Dispose (pf)
    end
end;

end.
